local RS = RingSlinger

local RELOADTIME = 2*TICRATE

local fireweapon = function(mo, weapon)
	S_StartSound(mo, sfx_rs_nor)
	local player = mo.player
	local mt = MT_RS_SHOT
	if weapon.mt
		mt = weapon.mt
	end
	
	if weapon.sound
		S_StartSound(mo, weapon.sound)
	end
	local shake = 0
	if weapon.shake
		shake = weapon.shake
	end
	if mo.ringslinger.powers[RSPOWER_POWERTOSS]
		shake = $ + 1
		S_StartSoundAtVolume(mo, sfx_kc68, 90)
	end
	if mo.ringslinger.powers[RSPOWER_RAPIDFIRE]
		S_StartSoundAtVolume(mo, sfx_kc4c, 60)
	end
	if mo.ringslinger.powers[RSPOWER_PIERCE]
		shake = $ + 2
		S_StartSoundAtVolume(mo, sfx_cdfm14, 100)
	end
	if shake
		if player == consoleplayer
			P_StartQuake(shake * FRACUNIT, min(5, shake + 2))
		end
	end
	if weapon.slingfunc
		weapon.slingfunc(mo, weapon)
		return
	end
	local shot = P_SPMAngle(mo, mt, mo.angle, 1, weapon.flags2)
	if shot and shot.valid
		if mo.ringslinger.powers[RSPOWER_POWERTOSS]
			shot.momx = $ * 3/2
			shot.momy = $ * 3/2
			shot.momz = $ * 3/2
		end
		if mo.ringslinger.powers[RSPOWER_PIERCE]
			shot.pierce = true
		end
		if weapon.flags2
			shot.flags2 = $ | weapon.flags2
		end
		shot.color = player.skincolor
		if weapon.scale
			shot.scale = FixedMul(weapon.scale, mo.scale)
		end
		if weapon.fuse
			shot.lifetime = weapon.fuse
		end
		if weapon.deathfunc
			shot.deathfunc = weapon.deathfunc
		end
	end
end

local tryfireweapon = function(mo)
	local player = mo.player
	local pbtn = player.rs_prevbuttons
	local btn = player.rs_buttons
	local weapon = RS.Weapons[mo.ringslinger.loadout[mo.ringslinger.wepslot]]
	local firebutton = (btn & BT_ATTACK) and (weapon.auto or not (pbtn & BT_ATTACK))
	local delay = weapon.delay
	local cost
	if weapon.cost != nil
		cost = weapon.cost
	else
		cost = 1
	end
	
	if (not firebutton) or mo.ringslinger.weapondelay
		return false
	end
	
	if mo.ringslinger.reload
		if not (pbtn & BT_ATTACK)
			S_StartSoundAtVolume(nil, sfx_kc44, 140, player)
		end
		return false
	end
	
	--Offhand penalty
	if mo.ringslinger.wepslot != 1
		if cost != 0
			cost = $ + max(1, $ / 2)
		end
		delay = $ + max(1, $ / 4)
	end
	
	--Rapid fire bonus
	if mo.ringslinger.powers[RSPOWER_RAPIDFIRE]
		delay = max(1, $ - max(1, $ / 3))
	end
	
	fireweapon(mo, weapon)
	
	--Ammo and delay
	mo.ringslinger.ammo = max(0, $ - cost)
	mo.ringslinger.lostrings = cost
	mo.ringslinger.lostringstimer = 18
	if mo.ringslinger.wepslot == 1
		mo.ringslinger.lostringsxoff = -10
	else
		mo.ringslinger.lostringsxoff = 10
	end
	mo.ringslinger.weapondelay = delay
	mo.ringslinger.boby = (6+weapon.delay)*FRACUNIT*7
	mo.ringslinger.swipe = 1
	if mo.ringslinger.ammo <= 0
		player.rings = max(0, $ - 20)
		mo.ringslinger.reloadcosthud = TICRATE/2
		mo.ringslinger.reload = RELOADTIME
		if (mo.ringslinger.powers[RSPOWER_QUICKRELOAD])
			mo.ringslinger.reload = $ * 1/3
		end
		S_StartSound(mo, sfx_antiri)
	end
end

local tryswap = function(mo)
	local player = mo.player
	local pbtn = player.rs_prevbuttons
	local btn = player.rs_buttons
	local swap = (btn & BT_FIRENORMAL) and not (pbtn & BT_FIRENORMAL)
	local nex = (btn & BT_WEAPONNEXT) and not (pbtn & BT_WEAPONNEXT)
	local prev = (btn & BT_WEAPONPREV) and not (pbtn & BT_WEAPONPREV)
	
	local newslot
	if swap
		if mo.ringslinger.wepslot == 1
			newslot = 2
		else
			newslot = 1
		end
	elseif nex
		newslot = 2
	elseif prev
		newslot = 1
	end
	
	if newslot
		if newslot == mo.ringslinger.wepslot
			S_StartSoundAtVolume(mo, sfx_s23a, 50)
			mo.ringslinger.hudy = 1
		else
			mo.ringslinger.wepslot = newslot
			mo.ringslinger.weapondelay = max($, 5)
			mo.ringslinger.boby = 10*FRACUNIT*7
			mo.ringslinger.hudy = 5
			S_StartSoundAtVolume(mo, sfx_s25d, 120)
			S_StartSoundAtVolume(mo, sfx_s23a, 70)
		end
	end
end

local numweapons = #RS.Weapons --This should not update after custom weapons are added.
local rsgetskin = function(skin)
	if RS.Skins[skin]
		return RS.Skins[skin]
	end
	
	local strbyte = string.byte(skin)
	local mainhand, offhand
	
	--All sonics will use scatter as their mainhand
	if string.match(skin, "sonic")
		mainhand = RSWPN_SCATTER
	end
	
	--Generate stuff based on skin's name and prefcolor
	if mainhand == nil
		mainhand = (strbyte % numweapons) + 1
	end
	if offhand == nil
		offhand = ((strbyte + skins[skin].prefcolor) % (numweapons - 1)) + 1
		if offhand >= mainhand
			offhand = $ + 1
		end
	end
	
	return {
		ammo = RS.defaultammo,
		weapons = {mainhand, offhand}
	}
end

addHook("PreThinkFrame", function()
	if not (gametyperules & GTR_RINGSLINGER) return end
	for player in players.iterate
		player.rs_buttons = player.cmd.buttons
		if player.mo
			player.cmd.buttons = $ & ~BT_ATTACK
			player.cmd.buttons = $ & ~BT_FIRENORMAL
		end
	end
end)

local resetplayer = function(mo, player)
	local rsskin = rsgetskin(mo.skin)
	mo.ringslinger = {}
	mo.ringslinger.skin = mo.skin
	mo.ringslinger.loadout = {rsskin.weapons[1], rsskin.weapons[2]}
	mo.ringslinger.wepslot = 1
	mo.ringslinger.weapondelay = TICRATE/2
	mo.ringslinger.ammo = rsskin.ammo
	mo.ringslinger.maxammo = rsskin.ammo
	mo.ringslinger.swipe = 0
	mo.ringslinger.bob = 0
	mo.ringslinger.bobx = 0
	mo.ringslinger.boby = FRACUNIT*120
	mo.ringslinger.hudy = 0
	mo.ringslinger.reload = 0
	mo.ringslinger.reloadcosthud = 0
	mo.ringslinger.lostringstimer = 0
	mo.ringslinger.lostrings = 0
	mo.ringslinger.lostringsxoff = 0
	mo.ringslinger.powers = {}
end

addHook("ThinkFrame", function()
	if not (gametyperules & GTR_RINGSLINGER) return end
	
	for player in players.iterate
		if not (player.mo and player.mo.valid and player.playerstate == PST_LIVE) continue end
		local mo = player.mo
		
		if (not mo.ringslinger)
			player.rs_prevbuttons = 0
			player.rs_buttons = 0
			resetplayer(mo, player)
			continue
		end
		
		if mo.skin != mo.ringslinger.skin
			resetplayer(mo, player)
			continue
		end
		
		--powerup timers
		for i = 1, #RS.Powers
			mo.ringslinger.powers[i] = $ or 0
			if mo.ringslinger.powers[i] > 0
				mo.ringslinger.powers[i] = $ - 1
				
				if RS.Powers[i].afterimages and (mo.ringslinger.powers[i] % 8 == 0)
					local g = P_SpawnGhostMobj(mo)
					g.frame = $ | FF_TRANS70
					g.destscale = $ * 2
					g.colorized = true
					g.color = RS.Powers[i].color
					P_Thrust(g, P_RandomRange(-50, 50)*ANG1 + mo.angle, P_RandomRange(-1, -6)*FRACUNIT)
				end
			end
		end
		
		--hud slide
		if mo.ringslinger.hudy
			mo.ringslinger.hudy = max(0, $ - 2)
		end
		
		--weapon delay
		if mo.ringslinger.weapondelay
			mo.ringslinger.weapondelay = $ - 1
		end
		
		--hurt
		if P_PlayerInPain(player)
			mo.ringslinger.weapondelay = 10
			mo.ringslinger.boby = 70*FRACUNIT
		end
		
		--lost ring display timer
		if mo.ringslinger.lostringstimer
			mo.ringslinger.lostringstimer = $ - 1
		end
		
		--reload
		if mo.ringslinger.reload > 0
			if mo.ringslinger.ammo > 0
				mo.ringslinger.reload = 0
			else
				mo.ringslinger.reload = $ - 1
				mo.ringslinger.boby = 100*FRACUNIT
				if mo.ringslinger.reload <= 0
					mo.ringslinger.ammo = mo.ringslinger.maxammo
					S_StartSoundAtVolume(nil, sfx_rs_amm, 120, mo.player)
				end
			end
		end
		if mo.ringslinger.reload == RELOADTIME - 12
			S_StartSound(mo, sfx_rs_rel)
		end
		
		--reload cost hud
		if mo.ringslinger.reloadcosthud
			mo.ringslinger.reloadcosthud = $ - 1
		end
		
		--bobbing
		mo.ringslinger.bob = min(FixedMul(player.rmomx,player.rmomx) + FixedMul(player.rmomy,player.rmomy) / 30, 7*FRACUNIT)
		local target
		if P_IsObjectOnGround(mo)
			target = FixedMul(sin(leveltime * ANG20), mo.ringslinger.bob)
		else
			target = max(min(10*FRACUNIT, mo.momz), -10*FRACUNIT)
		end
		local diff = max(target - mo.ringslinger.boby, -45*FRACUNIT)
		mo.ringslinger.boby = $ + diff / 4
		if P_IsObjectOnGround(mo)
			target = (-player.cmd.sidemove * FRACUNIT / 18)
		else
			target = 0
		end
		diff = target - mo.ringslinger.bobx
		mo.ringslinger.bobx = $ + diff / 10
		
		--swipe
		if mo.ringslinger.swipe
			mo.ringslinger.swipe = $ + 1
			if mo.ringslinger.swipe >= 4
				mo.ringslinger.swipe = 0
			end
		end
		
		--throwing rings, swapping weapons
		tryswap(mo)
		tryfireweapon(mo)
		
		player.rs_prevbuttons = player.rs_buttons
	end
end)

addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
	if (inflictor and inflictor.pierce and source and source.player and (target.player.powers[pw_shield] or target.ringslinger.powers[RSPOWER_PROTECTION]))
		P_RemoveShield(target.player)
		target.ringslinger.powers[RSPOWER_PROTECTION] = 0
		S_StartSoundAtVolume(nil, sfx_rs_pie, 150, target.player)
		S_StartSoundAtVolume(target, sfx_rs_pie, 150)
		S_StartSoundAtVolume(nil, sfx_rs_pie, 150, source.player)
		return
	end
	
	if (target.ringslinger and target.ringslinger.powers[RSPOWER_PROTECTION])
		S_StartSoundAtVolume(nil, sfx_cdfm28, 150, target.player)
		S_StartSoundAtVolume(target, sfx_cdfm28, 150)
		if source and source.player
			S_StartSoundAtVolume(nil, sfx_cdfm28, 150, source.player)
		end
		target.ringslinger.powers[RSPOWER_PROTECTION] = 0
		target.player.powers[pw_flashing] = max($, 25)
		target.momx = $ / 2
		target.momy = $ / 2
		if inflictor
			target.momx = $ + inflictor.momx / 3
			target.momy = $ + inflictor.momy / 3
			target.momz = $ + inflictor.momz / 3
		end
		return true
	end
end, MT_PLAYER)